#!/usr/bin/env python
# encoding: UTF-8
# api: python
# type: application
# title: streamtuner2
# description: directory browser for internet radio / audio streams
# depends: gtk, pygtk, xml.dom.minidom, threading
# version: 2.0.1
# author: mario
# license: public domain
# url: http://freshmeat.net/projects/streamtuner2
#
#
# Streamtuner2 is a GUI browser for internet radio directories. Various
# providers can be added, and streaming stations are usually grouped into
# music genres or categories.
#
# It's an independent rewrite of streamtuner1 in a scripting language. So
# it can be more easily extended and fixed. Feature paritiy still isn't
# there.
# Therefore streamtuner2 should be considered just an alternative to the
# original. It's not necessarily a replacement for everyone.
#
# Cumulative development time is just a few weeks now, but the application
# runs mostly stable now. The GUI interfaces are workable.
# There haven't been any optimizations regarding memory usage and
# performance. The current internal API is acceptable. Documentation is
# lacking.
#

""" project status """
#  current bugs:
#   - audio- and list-format support is not very robust / needs better API
#   - lots of GtkWarning messages
#   - not all keyboard shortcuts work
#   - in-list search doesn't work in our treeviews (???)
#   - still unfixed category reloading from live365
#   - JSON files are only trouble: loading of data files might lead to more
#     errors now, even if pson module still falls back on old method
#     (unicode strings from json.load are useless to us, require typecasts)
#     (nonsupport of tuples led to regression in mygtk.app_restore)
#
#  features:
#   - less functionality than original C streamtuner 0.99.x
#   - treeview lists are created from datamap[] structure and stream{} dicts
#   - channel categories are built-in defaults (can be freshened up however)
#   - config vars and cache data get stored as JSON in ~/.config/streamtuner2/
#
#  missing:
#   - localization
#
#  security notes:
#   - directory scrapers use fragile regular expressions - which is probably
#     not a security risk, but might lead to faulty data
#   - MEDIUM: little integrity checking for .pls / .m3u references and files
#   - minimal XML/SGML entity decoding (-> faulty data)
#   - MINIMAL: if system json module is not available, pseudo-json uses eval()
#     to read the config data -> limited risk, since it's only local files
#   - HIGH RISK: no verification of downloaded favicon image files (ico/png),
#     as they are passed to gtk.gdk.Pixbuf
#   - MEDIUM: audio players / decoders are easily affected by buffer overflows
#     from corrupt mp3/stream data, and streamtuner2 executes them
#      - but since that's the purpose -> no workaround
#
#  still help wanted on:
#   - any of the above
#   - new plugins (porting of basic.ch or punkcast, or local file viewer)
#   - nicer logo (or donations accepted to consult graphics designer)
#



# standard modules
import sys
import os
import re
import xml.dom.minidom
from xml.sax.saxutils import unescape as entity_decode, escape as xmlentities
import copy
import urllib

# gtk modules
import pygtk
import gtk
import gtk.glade
import gobject

# custom modules
sys.path.append("/usr/share/streamtuner2")   # pre-defined directory for modules
from processing import Process as Thread  # loads emulation with threading. on Python <= 2.5
from config import conf   # initializes itself, so all conf.vars are available right away
from mygtk import mygtk   # gtk treeview
import http
import action  # needs workaround... (action.main=main)
from channels import *
import favicon




# this represents the main window
# and also contains most application behaviour
class StreamTunerTwo(gtk.glade.XML):

        # constructor
        def __init__(self):

            # load gtk stylesheet
            if conf.get("theme"):
                for dir in (conf.dir, conf.share, "/usr/share"):
                    f = dir + "/themes/" + conf.theme + "/gtk-2.0/gtkrc"
                    if os.path.exists(f):
                        gtk.rc_parse(f)
                pass

            # instantiate gtk/glade widgets in current object
            gtk.glade.XML.__init__(self, ("st2.glade" if os.path.exists("st2.glade") else conf.share+"/st2.glade"))
                # gtk.glade.xml_new_from_buffer...
                
            # non-glade widgets (the manually instantiated ones)
            self.widgets = {}

            # widget shortcuts, helper vars
            self.sbar_cid = self.get_widget("statusbar").get_context_id("messages")
            self.sbar_msg = []
            # threads
            self.working = []

            # channel modules
            self.channels = {}  # needs to exist before instantiating modules, else any early __getattr__ access on self.channels[] creates race condition
            self.channels = {
              "bookmarks": bookmarks(parent=self),
              "shoutcast": shoutcast(parent=self),
              "live365": None,
              "xiph": None,
#              "search": None,
#              "google": None,
            }
            self.channel_names = ["bookmarks", "shoutcast", "live365", "xiph"]    # order of channel notebook tabs
            self.current_channel = "shoutcast"    # currently selected channel name (as index in self.channels{})
            
            # TEMPORARY ALIASES THAT WE DO NOT LIKE ANYMORE:
            #self.c = self.channels
            #self.cn = self.channel_names
            #self.cc = self.current_channel
            # YES, EVIL SHORTNAMES, YOU WILL BE REMOVED SOON
            
            # other channel modules / plugins
            self.load_plugin_channels()
            

            try:
                # load application state (widget sizes, selections, etc.)
                winlayout = conf.load("window")
                if (winlayout):
                    mygtk.app_restore(self, winlayout)
                # selection values
                winstate = conf.load("state")
                if (winstate):
                    for id in winstate.keys():
                        self.channels[id].current = winstate[id]["current"]
                        self.channels[id].shown = winlayout[id+"_list"].get("row:selected", 0)   # actually just used as boolean flag (for late loading of stream list), selection bar has been positioned before already
            except:
                pass

            # display current open channel/notebook tab
            self.current_channel = self.channel_names[self.notebook_channels.get_current_page()];
            self.channel().first_show()

      
            # bind gtk/glade event names to functions
            self.signal_autoconnect({
                "gtk_main_quit" : self.gtk_main_quit,                # close window
                # treeviews / notebook
                "on_stream_row_activated" : self.on_play_clicked,    # double click in a streams list
                "on_category_clicked": self.on_category_clicked,     # new selection in category list
                "on_notebook_channels_switch_page": self.channel_switch,   # channel notebook tab changed
                "station_context_menu": lambda tv,ev: station_context_menu(tv,ev),
                # toolbar
                "on_play_clicked" : self.on_play_clicked,
                "on_record_clicked": self.on_record_clicked,
                "on_homepage_stream_clicked": self.on_homepage_stream_clicked,
                "on_reload_clicked": self.on_reload_clicked,
                "on_stop_clicked": self.on_stop_clicked,
                "on_homepage_channel_clicked" : self.on_homepage_channel_clicked,
                "double_click_channel_tab": self.on_homepage_channel_clicked,
                # menu
                "menu_toolbar_standard": lambda(w): (self.toolbar.unset_style(), self.toolbar.unset_icon_size()),
                "menu_toolbar_style_icons": lambda(w): (self.toolbar.set_style(gtk.TOOLBAR_ICONS)),
                "menu_toolbar_style_both": lambda(w): (self.toolbar.set_style(gtk.TOOLBAR_BOTH)),
                "menu_toolbar_size_small": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_SMALL_TOOLBAR)),
                "menu_toolbar_size_medium": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_DND)),
                "menu_toolbar_size_large": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_DIALOG)),
                # else
                "menu_properties": config_dialog.open,
                "config_cancel": lambda(w): (self.win_config.hide()),
                "config_save": config_dialog.save,
                "update_categories": self.update_categories,
                "update_favicons": self.update_favicons,
                "app_state": self.app_state,
                "bookmark": self.bookmark,
                "save_as": self.save_as,
                "menu_about": lambda w: AboutStreamtuner2(),
                "menu_copy": self.menu_copy,
                "delete_entry": self.delete_entry,
                "search_open": search.menu_search,
                "search_set": search.search_set,
                "search_cancel": search.cancel,
                "streamedit_open": streamedit.open,
                "streamedit_save": streamedit.save,
                "streamedit_cancel": streamedit.cancel,
            })


            # actually display main window
            self.win_streamtuner2.show()
            
            # WHY DON'T YOU WANT TO WORK?!
            self.shoutcast.gtk_list.set_enable_search(True)
            self.shoutcast.gtk_list.set_search_column(3)

          

        # shortcut fo glade.get_widget(),
        # allows access to widgets as direct attributes instead of using .get_widget()
        # or looking in self.channels[] for the named channel plugins
        def __getattr__(self, name):
            if (self.channels.has_key(name)):
                return self.channels[name]
            else:
                return self.get_widget(name)
            # or gives an error, if neither exists
        def get_widget(self, name):
            if self.__dict__["widgets"].has_key(name):
                return self.__dict__["widgets"][name]
            else:
                #print "ST2.get_widget("+name+")\n"
                return gtk.glade.XML.get_widget(self, name)
                


                
        # returns the currently selected directory/channel object
        def channel(self):
            return self.channels[self.current_channel]

        # notebook tab clicked
        def channel_switch(self, notebook, page, page_num=0, *args):

            # can be called from channelmenu as well:
            if type(page) == str:
                self.current_channel = page
                self.notebook_channels.set_current_page(self.channel_names.index(page))
            # notebook invocation:
            else: #if type(page_num) == int:
                self.current_channel = self.channel_names[page_num]
            
            # if first selected, load current category
            self.channel().first_show()




        # convert ListStore iter to row number
        def rowno(self):
            (model, iter) = self.model_iter()
            return model.get_path(iter)[0]

        # currently selected entry in stations list, return complete data dict
        def row(self):
            return self.channel().stations() [self.rowno()]
            
        # return ListStore object and Iterator for currently selected row in gtk.TreeView station list
        def model_iter(self):
            return self.channel().gtk_list.get_selection().get_selected()
            
        # fetches a single varname from currently selected station entry
        def selected(self, name="url"):
            return self.row().get(name)

                


        # play button
        def on_play_clicked(self, widget, event=None, *more):
            row = self.row()
            print self.channel(), hasattr(self.channel(),"play")
            if hasattr(self.channel(),"play"):
                self.channel().play(row)
            else:
                action.play(url=row["url"], audioformat=row.get("format","audio/mp3"), listformat=self.channel().listformat)
            favicon.download_playing(self.row())

        # streamripper
        def on_record_clicked(self, widget):
            url = self.selected("url")
            action.record(url, "audio/mp3")

        # browse stream
        def on_homepage_stream_clicked(self, widget):
            url = self.selected("homepage")             
            action.browser(url)
             
        # browse channel
        def on_homepage_channel_clicked(self, widget, event=2):
            print "dblclick?"
            if event == 2 or event.type == gtk.gdk._2BUTTON_PRESS:
                action.browser(self.channel().homepage)            

        # reload stream list in current channel-category
        def on_reload_clicked(self, widget=None, reload=1):
            #print "reload", reload, self.current_channel, self.channels[self.current_channel], self.channel().current
            thread = Thread(target=self.channel().load, args=(self.channel().current, reload,))
            thread.start()
            self.working.append(thread)

        # stop reload/update threads
        def on_stop_clicked(self, widget):
            if self.working:
                thread = self.working.pop()
                thread.stop()
        
        # click in category list
        def on_category_clicked(self, widget, event, *more):
            category = self.channel().currentcat()
            #print "on_category_clicked", category, self.current_channel
            self.on_reload_clicked(None, reload=0)
            pass

        # add current selection to bookmark store
        def bookmark(self, widget):
            self.bookmarks.add(self.row())
            # code to update current list (set icon just in on-screen liststore, it would be updated with next display() anyhow - and there's no need to invalidate the ls cache, because that's referenced by model anyhow)
            try:
                (model,iter) = self.model_iter()
                model.set_value(iter, 0, gtk.STOCK_ABOUT)
            except:
                pass
            # refresh bookmarks tab
            self.bookmarks.load(self.bookmarks.default)

        # reload category tree
        def update_categories(self, widget):
            c = self.channel()
            # load from network, and save title list in cache/ file
            c.update_categories()
            conf.save("cache/categories_"+c.module, c.categories)
            # rebuild gtk.TreeView
            c.gtk_cat.remove_column(c.gtk_cat.get_column(0))
            mygtk.tree(c.gtk_cat, c.categories, title="Category", icon=gtk.STOCK_OPEN);
            # if it's a short list of categories, there's probably subfolders
            if len(c.categories) < 7:
                c.gtk_cat.expand_all()
            
        # menu invocation: refresh favicons for all stations in current streams category
        def update_favicons(self, widget):
            entries = self.channel().stations()
            favicon.download_all(entries)

        # save a file            
        def save_as(self, widget):
            row = self.row()
            default_fn = row["title"] + ".m3u"
            fn = mygtk.save_file("Save Stream", None, default_fn, [(".m3u","*m3u"),(".pls","*pls"),(".xspf","*xspf"),(".smil","*smil"),(".asx","*asx"),("all files","*")])
            if fn:
                action.save(row, fn)
            pass

        # save current stream URL into clipboard
        def menu_copy(self, w):
            gtk.clipboard_get().set_text(self.selected("url"))

        # remove an entry
        def delete_entry(self, w):
            n = self.rowno()
            del self.channel().stations()[ n ]
            self.channel().switch()
            self.channel().save()
            





        # shortcut to statusbar
        # (hacked to work from within threads, circumvents the statusbar msg pool actually)
        def status(self, text=""):
            # remove text
            while ((not text) and (type(text)==str) and len(self.sbar_msg)):
                self.sbar_msg.pop()
                mygtk.do(lambda:self.statusbar.pop(self.sbar_cid))
            # progressbar
            if (type(text)==float):
                if (text >= 1.0):  # completed
                    mygtk.do(lambda:self.progress.hide())
                else:  # show percentage
                    mygtk.do(lambda:self.progress.show() or self.progress.set_fraction(text))
                    if (text <= 0.0):  # unknown state
                        mygtk.do(lambda:self.progress.pulse())
            # add text
            elif (type(text)==str):
                self.sbar_msg.append(1)
                mygtk.do(lambda:self.statusbar.push(self.sbar_cid, text))
            pass

        # load plugins from /usr/share/streamtuner2/channels dir
        def load_plugin_channels(self):

            # find plugin files
            for fn in os.listdir(conf.share + "/channels/"):
                if (re.match("^[a-z0-9]+\.py$", fn)):
                    module = fn[0:-3] # cut filename
                    
                    # skip module if disabled
                    if conf.plugins.get(module, 1) == False:
                        print "disabled plugin:", module
                        continue
                    
                    # load plugin
                    plugin = __import__("channels."+module, None, None, [""])
                    plugin_class = plugin.__dict__[module]
                    
                    # load .config settings from plugin
                    conf.add_plugin_defaults(plugin_class.config, module)

                    # add and initialize object
                    if module not in self.channel_names:
                        self.channel_names.append(module)
                    self.channels[module] = plugin_class(parent=self)
            pass

        # store window/widget states (sizes, selections, etc.)
        def app_state(self, widget):
            # gtk widget states
            widgetnames = ["win_streamtuner2", "toolbar", "notebook_channels", ] \
                        + [id+"_list" for id in self.channel_names] + [id+"_cat" for id in self.channel_names]
            conf.save("window", mygtk.app_state(wTree=self, widgetnames=widgetnames))
            # object vars
            channelopts = {} #dict([(id, {"current":self.channels[id].current}) for id in self.channel_names])
            for id in self.channels.keys():
                if (self.channels[id]):
                    channelopts[id] = {"current":self.channels[id].current}
            conf.save("state", channelopts)

        # end application and gtk+ main loop
        def gtk_main_quit(self, widget):
            if conf.auto_save_appstate:
                self.app_state(widget)
            gtk.main_quit()

        def station_context_menu(self, treeview, event):
            return station_context_menu(treeview, event)





                




# auxiliary window: about dialog
class AboutStreamtuner2:
        # about us
        def __init__(self):
            a = gtk.AboutDialog()
            a.set_version("2.0.1")
            a.set_name("streamtuner2")
            a.set_license("Public Domain\n\nNo Strings Attached.\nUnrestricted distribution,\nmodification, use.")
            a.set_authors(["mario salzer <xmilky+st2@gm..>\n"])
            a.set_website("http://milki.erphesfurt.de/streamtuner2/")
            a.connect("response", lambda a, ok: ( a.hide(), a.destroy() ) )
            a.show()
            

            
# right click in streams/stations TreeView
def station_context_menu(treeview, event):
            # right-click ?
            if event.button >= 3:
                path = treeview.get_path_at_pos(event.x, event.y)[0]
                treeview.grab_focus()
                treeview.set_cursor(path, None, False)
                main.streamactions.popup(None, None, None, event.button, event.time)
                return None
            # we need to pass on to normal left-button signal handler
            else:
                return False
# this works better as callback function than as class - because of False/Object result for event trigger




# encapsulates references to gtk objects AND properties in main window
class auxiliary_window(object):
        def __getattr__(self, name):
            if main.__dict__.has_key(name):
                return main.__dict__[name]
            else:
                return main.get_widget(name)
""" allows to use self. and main. almost interchangably """



# aux win: search
class search (auxiliary_window):

        # show search dialog   
        def menu_search(self, w):
            self.search_dialog.show();


        # search text set
        def search_set(self, w):
            self.search_dialog.hide() #if conf.hide_searchdialog
            
            # save query string
            main.q = self.search_q.get_text().lower()

            # data attributes
            c = main.channel()
            rows = c.stations()
            col = c.rowmap.index("search_col")

            # callback to compare (+highlight) rows
            c.gtk_list.get_model().foreach(self.search_treestore, (rows, main.q, col, col+1))
            
            
            
        # callback that iterates over whole gtk treelist
        def search_treestore(self, model, path, iter, extra_data):
            i = path[0]
            (rows, q, color, flag) = extra_data

            # compare against content fields:
            text = rows[i].get("title") + " " + rows[i].get("homepage")
            text = text.lower()

            # just have a simple string match
            if text.find(q) >= 0:
               model.set_value(iter, color, "#fe9")  # highlighting color
               model.set_value(iter, flag, True) # background-set flag
            # color = 12 in liststore, flag = 13th position
            else:
               model.set_value(iter, color, "")   # for some reason the cellrenderer colors get applied to all rows, even if we specify an iter (=treelist position?, not?)
               model.set_value(iter, flag, False)   # that's why we need the secondary -set option

            #??
            return False


        def cancel(self, w):
            self.search_dialog.hide()

search = search()
# instantiates itself




# aux win: stream data editing dialog
class streamedit (auxiliary_window):

        # show stream data editing dialog
        def open(self, mw):
            row = main.row()
            for name in ("title", "playing", "genre", "homepage", "url", "favicon", "format", "extra"):
                w = main.get_widget("streamedit_" + name) 
                if w:
                    w.set_text((str(row.get(name)) if row.get(name) else ""))
            self.win_streamedit.show()

        # copy widget contents to stream
        def save(self, w):
            row = main.row()
            for name in ("title", "playing", "genre", "homepage", "url", "favicon", "format", "extra"):
               w = main.get_widget("streamedit_" + name)
               if w:
                   row[name] = w.get_text()
            main.channel().save()
            self.cancel(w)

        # hide window
        def cancel(self, w):
            self.win_streamedit.hide()

streamedit = streamedit()
# instantiates itself





# aux win: settings UI
class config_dialog (auxiliary_window):

        # display win_config, pre-fill text fields from global conf. object
        def open(self, widget):
            self.add_plugins()
            self.apply(conf.__dict__, "config_", 0)
            #self.win_config.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('#443399'))
            self.win_config.show()
        
        # set/load values between gtk window and conf. dict
        def apply(self, config, prefix="config_", save=0):
            for key,val in config.iteritems():
                # map non-alphanumeric chars from config{} to underscores in according gtk widget names
                id = re.sub("[^\w]", "_", key)
                w = main.get_widget(prefix + id)
           #print save, prefix+id, w, val
                # recurse into dictionaries, transform: conf.play.audio/mp3 => conf.play_audio_mp3
                if (type(val) == dict):
                    self.apply(val, prefix + id + "_", save)
                # load or set gtk.Entry text field
                elif (w and save and type(w)==gtk.Entry):
                    config[key] = w.get_text()
                elif (w and type(w)==gtk.Entry):
                    w.set_text(str(val))
                elif (w and save):
                    config[key] = w.get_active()
                elif (w):
                    w.set_active(bool(val))
            pass
            

        # add configuration setting definitions from plugins
        once = 0
        def add_plugins(self):
            if self.once:
                return

            for name,enabled in conf.plugins.iteritems():

                # add plugin load entry
                if name:
                    cb =  gtk.CheckButton(label="enable ⎗ "+name+" channel")
                    self.add_( "config_plugins_"+name, cb, label=None, color="#ddd" )

                # look up individual plugin options, if loaded
                if self.channels.get(name):
                    for opt in self.channels[name].config:

                        # add default values to conf. (now done in conf.add_plugin_defaults)
                        #if conf.__dict__.get(opt["name"]) == None:
                        #    print "set_default", opt
                        #    conf.__dict__[opt["name"]] = opt["value"]
                            
                        # display checkbox or text entry
                        if opt["type"] == "boolean":
                            self.add_( "config_"+opt["name"], gtk.CheckButton(opt["description"]) )
                        else:
                            self.add_( "config_"+opt["name"], gtk.Entry(), opt["description"] )

                # spacer 
                self.add_( "filler_pl_"+name, gtk.HSeparator() )
            self.once = 1

        # put gtk widgets into config dialog notebook
        def add_(self, id, w, label=None, color=""):
            w.set_property("visible", True)
            main.widgets[id] = w
            if label:
                w.set_width_chars(10)
                label = gtk.Label(label)
                label.set_property("visible", True)
                label.set_line_wrap(True) 
                label.set_size_request(250, -1)
                vbox = gtk.HBox(homogeneous=False, spacing=10)
                vbox.set_property("visible", True)
                vbox.pack_start(w, expand=False, fill=False)
                vbox.pack_start(label, expand=True, fill=True)
                w = vbox
            if color:
                w = mygtk.bg(w, color)
            self.plugin_options.pack_start(w)

        
        # save config
        def save(self, widget):
            self.apply(conf.__dict__, "config_", 1)
            conf.save()
            self.win_config.hide()
                  
config_dialog = config_dialog()
# instantiates itself





























# class GenericChannel:
#
#   is in channels/__init__.py
#





#-- favourite lists                                            ------------------------------------------
# this module lists static content from ~/.config/streamtuner2/bookmarks.json only,
# its data list is queries by other plugins to add 'star' icons
class bookmarks(GenericChannel):

        # desc
        api = "st2"
        module = "bookmarks"
        title = "bookmarks"
        version = 0.3
        base_url = "file:.config/streamtuner2/bookmarks.json"
        listformat = "*/*"
        
        # i like this
        config = [
            {"name":"like_my_bookmarks", "type":"boolean", "value":0, "description":"I like my bookmarks"},
        ]

        # content
        categories = ["favourite",]
        current = "favourite"
        default = "favourite"
        streams = {"favourite":[]}
        urls = []


        # this channel does not actually load any data
        def update_categories(self):
            pass
        def update_streams(self):
            pass
        def first_show(self):
            self.cache()
        # except from the cache file
        def cache(self):
            # stream list
            cache = conf.load(self.module)  # becomes "bookmarks.json"
            if (cache):
                self.streams = cache
            pass
        # save to cache file
        def save(self):
            conf.save(self.module, self.streams)
            pass


        # checks for existence of an URL in bookmarks store,
        # this method is called by other channel modules' display() method
        def is_in(self, url, once=1):
            if (not self.urls):
                self.urls = [row.get("url","urn:x-streamtuner2:no") for row in self.streams["favourite"]]
            return url in self.urls


        # called from main window / menu / context menu,
        # when bookmark is to be added for a selected stream entry
        def add(self, row):
        
            # normalize data (this row originated in a gtk+ widget)
            if row.get("favicon"):
               row["favicon"] = favicon.file(row.get("homepage"))
            row["favourite"] = 1
               
            # append to storage
            self.streams["favourite"].append(row)
            self.save()
            self.load(self.default)
            self.urls.append(row["url"])


        # simplified gtk TreeStore display logic (just one category for the moment, always rebuilt)
        def load(self, category, force=False):
            #self.liststore[category] = \
            mygtk.columns(self.gtk_list, self.datamap, self.display(self.streams[category]))
            pass












# SHOUTcast data module                                          ----------------------------------------
class shoutcast(GenericChannel):

        # desc
        api = "st2"
        module = "shoutcast"
        version = 0.9
        homepage = "http://www.shoutcast.com/" 
        base_url = "http://shoutcast.com/"
        listformat = "audio/x-scpls"

        # categories
        categories = ['default', 'Alternative', ['Adult Alternative', 'Britpop', 'Classic Alternative', 'College', 'Dancepunk', 'Dream Pop', 'Emo', 'Goth', 'Grunge', 'Hardcore', 'Indie Pop', 'Indie Rock', 'Industrial', 'Modern Rock', 'New Wave', 'Noise Pop', 'Power Pop', 'Punk', 'Ska', 'Xtreme'], 'Blues', ['Acoustic Blues', 'Chicago Blues', 'Contemporary Blues', 'Country Blues', 'Delta Blues', 'Electric Blues'], 'Classical', ['Baroque', 'Chamber', 'Choral', 'Classical Period', 'Early Classical', 'Impressionist', 'Modern', 'Opera', 'Piano', 'Romantic', 'Symphony'], 'Country', ['Americana', 'Bluegrass', 'Classic Country', 'Contemporary Bluegrass', 'Contemporary Country', 'Honky Tonk', 'Hot Country Hits', 'Western'], 'Decades', ['30s', '40s', '50s', '60s', '70s', '80s', '90s'], 'Easy Listening', ['Exotica', 'Light Rock', 'Lounge', 'Orchestral Pop', 'Polka', 'Space Age Pop'], 'Electronic', ['Acid House', 'Ambient', 'Big Beat', 'Breakbeat', 'Dance', 'Demo', 'Disco', 'Downtempo', 'Drum and Bass', 'Electro', 'Garage', 'Hard House', 'House', 'IDM', 'Jungle', 'Progressive', 'Techno', 'Trance', 'Tribal', 'Trip Hop'], 'Folk', ['Alternative Folk', 'Contemporary Folk', 'Folk Rock', 'New Acoustic', 'Traditional Folk', 'World Folk'], 'Inspirational', ['Christian', 'Christian Metal', 'Christian Rap', 'Christian Rock', 'Classic Christian', 'Contemporary Gospel', 'Gospel', 'Southern Gospel', 'Traditional Gospel'], 'International', ['African', 'Arabic', 'Asian', 'Bollywood', 'Brazilian', 'Caribbean', 'Celtic', 'Chinese', 'European', 'Filipino', 'French', 'Greek', 'Hindi', 'Indian', 'Japanese', 'Jewish', 'Klezmer', 'Korean', 'Mediterranean', 'Middle Eastern', 'North American', 'Russian', 'Soca', 'South American', 'Tamil', 'Worldbeat', 'Zouk'], 'Jazz', ['Acid Jazz', 'Avant Garde', 'Big Band', 'Bop', 'Classic Jazz', 'Cool Jazz', 'Fusion', 'Hard Bop', 'Latin Jazz', 'Smooth Jazz', 'Swing', 'Vocal Jazz', 'World Fusion'], 'Latin', ['Bachata', 'Banda', 'Bossa Nova', 'Cumbia', 'Latin Dance', 'Latin Pop', 'Latin Rock', 'Mariachi', 'Merengue', 'Ranchera', 'Reggaeton', 'Regional Mexican', 'Salsa', 'Tango', 'Tejano', 'Tropicalia'], 'Metal', ['Black Metal', 'Classic Metal', 'Extreme Metal', 'Grindcore', 'Hair Metal', 'Heavy Metal', 'Metalcore', 'Power Metal', 'Progressive Metal', 'Rap Metal'], 'Misc', [], 'New Age', ['Environmental', 'Ethnic Fusion', 'Healing', 'Meditation', 'Spiritual'], 'Pop', ['Adult Contemporary', 'Barbershop', 'Bubblegum Pop', 'Dance Pop', 'Idols', 'JPOP', 'Oldies', 'Soft Rock', 'Teen Pop', 'Top 40', 'World Pop'], 'Public Radio', ['College', 'News', 'Sports', 'Talk'], 'Rap', ['Alternative Rap', 'Dirty South', 'East Coast Rap', 'Freestyle', 'Gangsta Rap', 'Hip Hop', 'Mixtapes', 'Old School', 'Turntablism', 'West Coast Rap'], 'Reggae', ['Contemporary Reggae', 'Dancehall', 'Dub', 'Ragga', 'Reggae Roots', 'Rock Steady'], 'Rock', ['Adult Album Alternative', 'British Invasion', 'Classic Rock', 'Garage Rock', 'Glam', 'Hard Rock', 'Jam Bands', 'Piano Rock', 'Prog Rock', 'Psychedelic', 'Rockabilly', 'Surf'], 'Soundtracks', ['Anime', 'Kids', 'Original Score', 'Showtunes', 'Video Game Music'], 'Talk', ['BlogTalk', 'Comedy', 'Community', 'Educational', 'Government', 'News', 'Old Time Radio', 'Other Talk', 'Political', 'Scanner', 'Spoken Word', 'Sports', 'Technology'], 'Themes', ['Adult', 'Best Of', 'Chill', 'Eclectic', 'Experimental', 'Female', 'Heartache', 'Instrumental', 'LGBT', 'Party Mix', 'Patriotic', 'Rainy Day Mix', 'Reality', 'Sexy', 'Shuffle', 'Travel Mix', 'Tribute', 'Trippy', 'Work Mix']]
        #["default", [], 'TopTen', [], 'Alternative', ['College', 'Emo', 'Hardcore', 'Industrial', 'Punk', 'Ska'], 'Americana', ['Bluegrass', 'Blues', 'Cajun', 'Folk'], 'Classical', ['Contemporary', 'Opera', 'Symphonic'], 'Country', ['Bluegrass', 'New Country', 'Western Swing'], 'Electronic', ['Acid Jazz', 'Ambient', 'Breakbeat', 'Downtempo', 'Drum and Bass', 'House', 'Trance', 'Techno'], 'Hip Hop', ['Alternative', 'Hardcore', 'New School', 'Old School', 'Turntablism'], 'Jazz', ['Acid Jazz', 'Big Band', 'Classic', 'Latin', 'Smooth', 'Swing'], 'Pop/Rock', ['70s', '80s', 'Classic', 'Metal', 'Oldies', 'Pop', 'Rock', 'Top 40'], 'R&B/Soul', ['Classic', 'Contemporary', 'Funk', 'Smooth', 'Urban'], 'Spiritual', ['Alternative', 'Country', 'Gospel', 'Pop', 'Rock'], 'Spoken', ['Comedy', 'Spoken Word', 'Talk'], 'World', ['African', 'Asian', 'European', 'Latin', 'Middle Eastern', 'Reggae'], 'Other/Mixed', ['Eclectic', 'Film', 'Instrumental']]
        current = ""
        default = "default"
        empty = "default"
        
        # redefine
        streams = {}
        
            
        # extracts the category list from shoutcast.com,
        # the text/html format always being:
        #        <OPTION VALUE="Classical">Classical
        #                <OPTION VALUE="Contemporary"> - Contemporary
        #                <OPTION VALUE="Opera"> - Opera
        # sub-categories are indented by two tabs, maincats just by one
        def update_categories(self):
            html = http.get(self.base_url)
            self.categories = ["default"]
            print html

            # <h2>Radio Genres</h2>
	    rx_main = re.compile(r'<li class="prigen" id="(\d+)".+?<a href="/radio/([\w\s]+)">[\w\s]+</a></li>', re.S)
	    rx_sub = re.compile(r'<a href="/radio/([\w\s\d]+)">[\w\s\d]+</a></li>')
            for uu in rx_main.findall(html):
                print uu
		(id,name) = uu
                name = urllib.unquote(name)

                # main category
                self.categories.append(name)

                # sub entries
                html = http.ajax("http://shoutcast.com/genre.jsp", {"genre":name, "id":id})
                print html
                sub = rx_sub.findall(html)
                self.categories.append(sub)

            # it's done
            print self.categories
            conf.save("cache/categories_shoutcast", self.categories)
            pass



        def strip_tags(self, s):
            rx = re.compile(""">(\w+)<""")
            return " ".join(rx.findall(s))

        # downloads stream list from shoutcast for given category
        def update_streams(self, cat, search=""):

            if (not cat or cat == self.empty):
                print "nocat"
                return []
            ucat = urllib.quote(cat)

            
            # new extraction regex
            rx_stream = re.compile("""
                 <a\s+class="?playbutton[^>]+id="(\d+)".+?
		 target="_blank"\s+href="(http://[^">]+)"[^>]*>([^<>]+)</a>.+?
                 Recently\s*played:\s*([^<]*).+?
                 ners">(\d*)<.+?
                 bitrate">(\d*)<.+?
                 type">([MP3AAC]*)
                 """, re.S|re.I|re.X)
            rx_next = re.compile("""onclick="showMoreGenre""")
            # @meh: it's time for some jquery here



            # loop
            entries = []
            next = 0
            max = int(conf.max_streams)
            count = max
            while (next < max):

               # page
               url = "http://www.shoutcast.com/genre-ajax/" + ucat
	       referer = url.replace("/genre-ajax", "/radio")
	       params = { "strIndex":str(0),  "count":str(count), "ajax":"true", "mode":"listeners", "order":"desc"}
               html = http.ajax(url, params, referer)   #,feedback=self.parent.status)

               print html

               
               # extract entries
               self.parent.status("parsing document...")
              # print "loop-rx"
               for uu in rx_stream.findall(html):
		   (id, homepage, title, playing, ls, bit, fmt) = uu
                   print uu
                   entries += [{
                       "title": entity_decode(title),
                       "url": "http://yp.shoutcast.com/sbin/tunein-station.pls?id=" + id,
                       "homepage": http.fix_url(homepage),
                       "playing": entity_decode(playing),
                       "genre": cat, #self.strip_tags(uu[4]),
                       "listeners": int(ls),
                       "max": 0, #int(uu[6]),
                       "bitrate": int(bit),
                       "format": "audio/" + fmt.lower(),
                   }]

              # print "update_windows"
               self.parent.status()
               self.update_streams_partially_done(entries)
              # print "done"
               
               # more pages to load?
               if (re.search(rx_next, html)):
                  next += count
               else:
                  next = 99999
            
            #fin
            return entries

















#-- run main                                ---------------------------------------------
if __name__ == "__main__":

    #-- global configuration settings
    "conf = Config()"
    # already happened with "from config import conf"
    
    # prepare for threading in Gtk+ callbacks
    gobject.threads_init()
    
    # prepare main window
    main = StreamTunerTwo()
    
    # action (play/record) module needs a reference to main window
    action.main = main      # for gtk interaction and some URL/URI callbacks
    action = action.action  # shorter name
    
    # first invocation
    if (conf.__dict__.get("firstrun")):
        config_dialog.open(None)
        del conf.firstrun

    # run
    gtk.main()
    



#
#
#
#
